import uuid
from unittest.mock import patch

from django.urls import reverse
from rest_framework import status

from arkindex.process.models import ProcessMode, WorkerRun, WorkerVersion
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role


class TestBulkTranscriptionEntities(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.worker_version = WorkerVersion.objects.get(worker__slug="reco")
        cls.worker_run = cls.worker_version.worker_runs.filter(process__mode=ProcessMode.Workers).get()
        cls.custom_version = WorkerVersion.objects.get(worker__slug="custom")
        cls.local_worker_run = WorkerRun.objects.get(version=cls.custom_version, process__creator=cls.user)
        cls.transcription = cls.corpus.elements.get(name="Volume 1, page 1r").transcriptions.create(
            text="Once upon a time in a castle a knight was living lonely",
            confidence=0.42,
        )
        cls.location_ent_type = cls.corpus.entity_types.get(name="location")
        cls.person_ent_type = cls.corpus.entity_types.get(name="person")
        cls.entity = cls.corpus.entities.create(name="Paris", type=cls.location_ent_type)

    def test_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(reverse("api:transcription-entities-bulk", kwargs={"pk": self.transcription.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.post(reverse("api:transcription-entities-bulk", kwargs={"pk": self.transcription.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    @patch("arkindex.project.mixins.has_access", return_value=False)
    def test_requires_contributor(self, has_access_mock):
        self.user.rights.update(level=Role.Guest.value)
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(reverse("api:transcription-entities-bulk", kwargs={"pk": self.transcription.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(response.json(), {
            "detail": "You do not have a contributor access to the corpus of this transcription."
        })

    def test_not_found(self):
        self.client.force_login(self.superuser)
        with self.assertNumQueries(3):
            response = self.client.post(reverse("api:transcription-entities-bulk", kwargs={"pk": uuid.uuid4()}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_required_fields(self):
        self.client.force_login(self.superuser)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:transcription-entities-bulk", kwargs={"pk": str(self.transcription.id)}),
                data={},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "entities": ["This field is required."],
            "worker_run_id": ["This field is required."],
        })

    def test_wrong_values(self):
        fake_uuid = uuid.uuid4()
        self.client.force_login(self.superuser)
        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:transcription-entities-bulk", kwargs={"pk": str(self.transcription.id)}),
                data={
                    "entities": [],
                    "worker_run_id": str(fake_uuid),
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "entities": {"non_field_errors": ["This list may not be empty."]},
            "worker_run_id": [f'Invalid pk "{fake_uuid}" - object does not exist.'],
        })

    def test_existing_transcription_entity(self):
        self.transcription.transcription_entities.create(
            entity=self.entity,
            offset=1,
            length=1,
            worker_run=self.local_worker_run,
            worker_version=self.local_worker_run.version,
        )
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:transcription-entities-bulk", kwargs={"pk": str(self.transcription.id)}),
                data={
                    "entities": [
                        {
                            "name": "Knight",
                            "type_id": str(self.person_ent_type.id),
                            "offset": 10,
                            "length": 5,
                            "confidence": 0.05,
                        },
                    ],
                    "worker_run_id": str(self.local_worker_run.id),
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "worker_run_id": ["Some TranscriptionEntities were already created for this worker run on this transcription."],
        })

    def test_entity_fields_validation(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:transcription-entities-bulk", kwargs={"pk": str(self.transcription.id)}),
                data={
                    "entities": [
                        {}
                    ],
                    "worker_run_id": str(self.local_worker_run.id),
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "entities": [{
                "length": ["This field is required."],
                "name": ["This field is required."],
                "offset": ["This field is required."],
                "type_id": ["This field is required."],
            }]
        })

    def test_duplicate_entities(self):
        self.client.force_login(self.user)
        payload = {
            "name": "Knight",
            "type_id": str(self.person_ent_type.id),
            "offset": 10,
            "length": 5,
            "confidence": 0.05,
        }
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:transcription-entities-bulk", kwargs={"pk": str(self.transcription.id)}),
                data={
                    "entities": [
                        payload,
                        payload,
                    ],
                    "worker_run_id": str(self.local_worker_run.id),
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "entities": {
                "__all__": ["Some TranscriptionEntities have the same name, type, offset and length."],
            },
        })

    def test_create_non_local(self):
        """
        A regular user cannot create transcription's entities with a WorkerRun of a non-local process
        """
        self.client.force_login(self.superuser)
        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:transcription-entities-bulk", kwargs={"pk": str(self.transcription.id)}),
                data={
                    "entities": [
                        {
                            "name": "Vileplume",
                            "type_id": str(self.person_ent_type.id),
                            "offset": 6,
                            "length": 9,
                            "confidence": .42,
                        },
                    ],
                    "worker_run_id": str(self.worker_run.id),
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "worker_run_id": [
                "Ponos task authentication is required to use a WorkerRun "
                "of a process other than the user's local process."
            ]
        })

    def test_create_other_user(self):
        """
        A regular user cannot create transcription's entities with a WorkerRun of someone else's local process
        """
        private_worker_run = self.user.processes.get(mode=ProcessMode.Local).worker_runs.first()
        self.client.force_login(self.superuser)
        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:transcription-entities-bulk", kwargs={"pk": str(self.transcription.id)}),
                data={
                    "entities": [
                        {
                            "name": "Vileplume",
                            "type_id": str(self.person_ent_type.id),
                            "offset": 6,
                            "length": 9,
                            "confidence": .42,
                        },
                    ],
                    "worker_run_id": str(private_worker_run.id),
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "worker_run_id": [
                "Ponos task authentication is required to use a WorkerRun "
                "of a process other than the user's local process."
            ]
        })

    def test_create_other_process(self):
        """
        A Ponos task cannot create transcription's entities with a WorkerRun of another process
        """
        process2 = self.worker_run.process.creator.processes.create(
            mode=ProcessMode.Workers,
            corpus=self.corpus,
        )
        other_worker_run = process2.worker_runs.create(version=self.worker_run.version, parents=[])
        self.worker_run.process.run()

        task = self.worker_run.process.tasks.first()
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:transcription-entities-bulk", kwargs={"pk": str(self.transcription.id)}),
                data={
                    "entities": [
                        {
                            "name": "Guzzlord",
                            "type_id": str(self.person_ent_type.id),
                            "offset": 6,
                            "length": 9,
                            "confidence": .42,
                        },
                    ],
                    "worker_run_id": str(other_worker_run.id),
                },
                HTTP_AUTHORIZATION=f"Ponos {task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "worker_run_id": [
                "Only the WorkerRuns of the authenticated task's process may be used."
            ]
        })

    def test_create_local(self):
        """
        A regular user can create entities with their links to a
        trancription using a WorkerRun of their own local process.
        """
        self.client.force_login(self.user)
        self.assertQuerySetEqual(
            self.corpus.entities.values_list("name", "type", "worker_run_id").order_by("name", "type", "worker_run_id"),
            [("Paris", self.location_ent_type.id, None)],
        )
        self.assertEqual(self.transcription.transcription_entities.count(), 0)

        with self.assertNumQueries(10):
            response = self.client.post(
                reverse("api:transcription-entities-bulk", kwargs={"pk": str(self.transcription.id)}),
                data={
                    "entities": [
                        {
                            # This entity already exists, and should be duplicated
                            "name": "Paris",
                            "type_id": str(self.location_ent_type.id),
                            "offset": 0,
                            "length": 1,
                            "confidence": 0.7,
                        },
                        {
                            "name": "Knight",
                            "type_id": str(self.person_ent_type.id),
                            "offset": 0,
                            "length": 1,
                            "confidence": None,
                        },
                        {
                            "name": "Knight",
                            "type_id": str(self.person_ent_type.id),
                            "offset": 10,
                            "length": 5,
                        },
                    ],
                    "worker_run_id": str(self.local_worker_run.id),
                },
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertQuerySetEqual(
            self.corpus.entities.values_list("name", "type", "worker_run_id").order_by("name", "type", "worker_run_id"),
            [
                ("Knight", self.person_ent_type.id, self.local_worker_run.id),
                ("Knight", self.person_ent_type.id, self.local_worker_run.id),
                ("Paris", self.location_ent_type.id, self.local_worker_run.id),
                ("Paris", self.location_ent_type.id, None),
            ],
        )

        self.assertQuerySetEqual(
            self.transcription.transcription_entities
            .values_list("entity__name", "entity__type", "offset", "length", "confidence", "worker_run_id")
            .order_by("entity__name", "offset"),
            [
                ("Knight", self.person_ent_type.id, 0, 1, None, self.local_worker_run.id),
                ("Knight", self.person_ent_type.id, 10, 5, None, self.local_worker_run.id),
                ("Paris", self.location_ent_type.id, 0, 1, 0.7, self.local_worker_run.id),
            ],
        )

        knight1, knight2, paris = self.transcription.transcription_entities.order_by("entity__name", "offset")

        # Ensure that the returned IDs match the order in which the payloads were sent
        self.assertEqual(response.json(), {
            "entities": [
                {
                    "transcription_entity_id": str(paris.id),
                    "entity_id": str(paris.entity_id),
                },
                {
                    "transcription_entity_id": str(knight1.id),
                    "entity_id": str(knight1.entity_id),
                },
                {
                    "transcription_entity_id": str(knight2.id),
                    "entity_id": str(knight2.entity_id),
                },
            ],
        })

    def test_create_task_auth(self):
        """
        Transcription's entities can be created with a WorkerRun of a non-local process
        when authenticated as a Ponos task of this process
        """
        self.worker_run.process.run()
        task = self.worker_run.process.tasks.first()

        with self.assertNumQueries(9):
            response = self.client.post(
                reverse("api:transcription-entities-bulk", kwargs={"pk": str(self.transcription.id)}),
                data={
                    "entities": [
                        {
                            "name": "Magmortar",
                            "type_id": str(self.person_ent_type.id),
                            "offset": 6,
                            "length": 9,
                            "confidence": .42,
                        },
                    ],
                    "worker_run_id": str(self.worker_run.id),
                },
                HTTP_AUTHORIZATION=f"Ponos {task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertQuerySetEqual(
            self.transcription.transcription_entities.values_list(
                "entity__name", "entity__type", "offset", "length", "confidence", "worker_run_id", "entity__worker_run_id"
            ),
            [("Magmortar", self.person_ent_type.id, 6, 9, .42, self.worker_run.id, self.worker_run.id)],
        )

    def test_create_local_task_auth(self):
        """
        Transcription's entities can be created with a WorkerRun of a Local process
        even when authenticated as a Ponos task from a different process
        """
        self.worker_run.process.run()
        task = self.worker_run.process.tasks.first()

        with self.assertNumQueries(9):
            response = self.client.post(
                reverse("api:transcription-entities-bulk", kwargs={"pk": str(self.transcription.id)}),
                data={
                    "entities": [
                        {
                            "name": "Guzzlord",
                            "type_id": str(self.person_ent_type.id),
                            "offset": 6,
                            "length": 9,
                            "confidence": .42,
                        },
                    ],
                    "worker_run_id": str(self.local_worker_run.id),
                },
                HTTP_AUTHORIZATION=f"Ponos {task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertQuerySetEqual(
            self.transcription.transcription_entities.values_list(
                "entity__name", "entity__type", "offset", "length", "confidence", "worker_run_id", "entity__worker_run_id"
            ),
            [("Guzzlord", self.person_ent_type.id, 6, 9, .42, self.local_worker_run.id, self.local_worker_run.id)],
        )
