from io import StringIO
from textwrap import dedent
from unittest.mock import patch

from django.core.management import CommandError, call_command
from django.utils import timezone

from arkindex.process.models import ArkindexFeature, FeatureUsage, Worker, WorkerType, WorkerVersion, WorkerVersionState
from arkindex.project.tests import ArkindexTestCase

MOCK_CONFIG = {
    "features": {
        "init_elements": "registry.example.com/init-elements:latest",
        "s3_ingest": "registry.example.com/s3-ingest:latest",
    }
}


class TestUpdateSystemWorkers(ArkindexTestCase):

    def update_system_workers(self):
        output = StringIO()
        call_command(
            "update_system_workers",
            ["--no-color"],
            stdout=output,
            stderr=output,
        )
        return output.getvalue().strip()

    def check_feature_version(self, worker_version, feature):
        """
        Assert that a WorkerVersion has the expected attributes to provide a given feature
        """
        self.assertEqual(worker_version.feature, feature)
        self.assertEqual(worker_version.docker_image_iid, MOCK_CONFIG["features"][feature.value])
        self.assertEqual(worker_version.state, WorkerVersionState.Available)
        self.assertNotIn("command", worker_version.configuration.get("docker", {}))
        self.assertEqual(worker_version.required_user_configuration_fields, set())
        self.assertEqual(worker_version.gpu_usage, FeatureUsage.Disabled)
        self.assertEqual(worker_version.model_usage, FeatureUsage.Disabled)

        self.assertIsNone(worker_version.worker.archived)
        self.assertTrue(worker_version.worker.public)

    @patch("arkindex.process.management.commands.update_system_workers.parse_config", lambda: MOCK_CONFIG)
    def test_from_scratch(self):
        output = self.update_system_workers()

        worker_type = WorkerType.objects.get()
        self.assertEqual(worker_type.slug, "system")
        self.assertEqual(worker_type.display_name, "System")

        init_elements_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.InitElements)
        self.check_feature_version(init_elements_version, ArkindexFeature.InitElements)

        s3_ingest_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.S3Ingest)
        self.check_feature_version(s3_ingest_version, ArkindexFeature.S3Ingest)

        # Check the attributes on the two new workers
        self.assertEqual(init_elements_version.worker.name, "InitElements")
        self.assertEqual(init_elements_version.worker.slug, "init_elements")
        self.assertEqual(init_elements_version.worker.type, worker_type)
        self.assertEqual(s3_ingest_version.worker.name, "S3Ingest")
        self.assertEqual(s3_ingest_version.worker.slug, "s3_ingest")
        self.assertEqual(s3_ingest_version.worker.type, worker_type)

        self.assertEqual(output, dedent(
            f"""
            ───────────────────────────────── InitElements ─────────────────────────────────
            Using registry.example.com/init-elements:latest to provide InitElements
            Created new System worker type ({worker_type.id})
            Created new InitElements system worker
            Creating new worker version
            Using new worker version {init_elements_version.id}
            ─────────────────────────────────── S3Ingest ───────────────────────────────────
            Using registry.example.com/s3-ingest:latest to provide S3Ingest
            Created new S3Ingest system worker
            Creating new worker version
            Using new worker version {s3_ingest_version.id}
            """
        ).strip())

    @patch("arkindex.process.management.commands.update_system_workers.parse_config", lambda: MOCK_CONFIG)
    def test_existing_worker_type(self):
        worker_type = WorkerType.objects.create(slug="system", display_name="Système")

        output = self.update_system_workers()

        self.assertEqual(WorkerType.objects.count(), 1)
        init_elements_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.InitElements)
        self.check_feature_version(init_elements_version, ArkindexFeature.InitElements)

        s3_ingest_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.S3Ingest)
        self.check_feature_version(s3_ingest_version, ArkindexFeature.S3Ingest)

        # Check the attributes on the two new workers
        self.assertEqual(init_elements_version.worker.name, "InitElements")
        self.assertEqual(init_elements_version.worker.slug, "init_elements")
        self.assertEqual(init_elements_version.worker.type, worker_type)
        self.assertEqual(s3_ingest_version.worker.name, "S3Ingest")
        self.assertEqual(s3_ingest_version.worker.slug, "s3_ingest")
        self.assertEqual(s3_ingest_version.worker.type, worker_type)

        self.assertEqual(output, dedent(
            f"""
            ───────────────────────────────── InitElements ─────────────────────────────────
            Using registry.example.com/init-elements:latest to provide InitElements
            Created new InitElements system worker
            Creating new worker version
            Using new worker version {init_elements_version.id}
            ─────────────────────────────────── S3Ingest ───────────────────────────────────
            Using registry.example.com/s3-ingest:latest to provide S3Ingest
            Created new S3Ingest system worker
            Creating new worker version
            Using new worker version {s3_ingest_version.id}
            """
        ).strip())

    @patch("arkindex.process.management.commands.update_system_workers.parse_config", lambda: MOCK_CONFIG)
    def test_existing_system_workers(self):
        worker_type = WorkerType.objects.create(slug="system", display_name="Système")
        init_elements_worker = Worker.objects.create(
            type=worker_type,
            slug="init_elements",
            name="Init of the Elements",
            public=True,
        )
        s3_ingest_worker = Worker.objects.create(
            type=worker_type,
            slug="s3_ingest",
            name="S3, Ingest of the",
            public=True,
        )

        output = self.update_system_workers()

        self.assertEqual(WorkerType.objects.count(), 1)
        self.assertEqual(Worker.objects.count(), 2)

        init_elements_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.InitElements)
        self.assertEqual(init_elements_version.worker, init_elements_worker)
        self.check_feature_version(init_elements_version, ArkindexFeature.InitElements)

        s3_ingest_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.S3Ingest)
        self.assertEqual(s3_ingest_version.worker, s3_ingest_worker)
        self.check_feature_version(s3_ingest_version, ArkindexFeature.S3Ingest)

        self.assertEqual(output, dedent(
            f"""
            ───────────────────────────────── InitElements ─────────────────────────────────
            Using registry.example.com/init-elements:latest to provide InitElements
            Using existing system worker Init of the Elements
            Worker is up to date
            Creating new worker version
            Using new worker version {init_elements_version.id}
            ─────────────────────────────────── S3Ingest ───────────────────────────────────
            Using registry.example.com/s3-ingest:latest to provide S3Ingest
            Using existing system worker S3, Ingest of the
            Worker is up to date
            Creating new worker version
            Using new worker version {s3_ingest_version.id}
            """
        ).strip())

    @patch("arkindex.process.management.commands.update_system_workers.parse_config", lambda: MOCK_CONFIG)
    def test_worker_slug_conflict(self):
        worker_type = WorkerType.objects.create(slug="systemnt", display_name="Systemn't")

        worker = Worker.objects.create(
            type=worker_type,
            slug="init_elements",
            name="Troublemaker",
        )

        output = StringIO()
        with self.assertRaisesMessage(CommandError, "Cannot create system worker with slug init_elements due to a conflict."):
            call_command(
                "update_system_workers",
                ["--no-color"],
                stdout=output,
                stderr=output,
            )

        self.assertEqual(output.getvalue().strip(), dedent(
            f"""
            ───────────────────────────────── InitElements ─────────────────────────────────
            Using registry.example.com/init-elements:latest to provide InitElements
            Worker Troublemaker ({worker.id}) uses the same slug with a different worker type.
            Please update this worker's type to `system` or change its slug.
            """
        ).strip())

    @patch("arkindex.process.management.commands.update_system_workers.parse_config", lambda: MOCK_CONFIG)
    def test_update_existing_system_workers(self):
        worker_type = WorkerType.objects.create(slug="system", display_name="Système")
        init_elements_worker = Worker.objects.create(
            type=worker_type,
            slug="init_elements",
            name="Init of the Elements",
            # This worker will be made public
            public=False,
        )
        s3_ingest_worker = Worker.objects.create(
            type=worker_type,
            slug="s3_ingest",
            name="S3, Ingest of the",
            public=True,
            # This worker will be unarchived
            archived=timezone.now(),
        )

        output = self.update_system_workers()

        self.assertEqual(WorkerType.objects.count(), 1)
        self.assertEqual(Worker.objects.count(), 2)

        init_elements_worker.refresh_from_db()
        self.assertEqual(init_elements_worker.name, "Init of the Elements")
        self.assertEqual(init_elements_worker.slug, "init_elements")
        self.assertEqual(init_elements_worker.type, worker_type)
        self.assertTrue(init_elements_worker.public)
        self.assertIsNone(init_elements_worker.archived)

        s3_ingest_worker.refresh_from_db()
        self.assertEqual(s3_ingest_worker.name, "S3, Ingest of the")
        self.assertEqual(s3_ingest_worker.slug, "s3_ingest")
        self.assertEqual(s3_ingest_worker.type, worker_type)
        self.assertTrue(s3_ingest_worker.public)
        self.assertIsNone(s3_ingest_worker.archived)

        init_elements_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.InitElements)
        self.assertEqual(init_elements_version.worker, init_elements_worker)
        self.check_feature_version(init_elements_version, ArkindexFeature.InitElements)

        s3_ingest_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.S3Ingest)
        self.assertEqual(s3_ingest_version.worker, s3_ingest_worker)
        self.check_feature_version(s3_ingest_version, ArkindexFeature.S3Ingest)

        self.assertEqual(output, dedent(
            f"""
            ───────────────────────────────── InitElements ─────────────────────────────────
            Using registry.example.com/init-elements:latest to provide InitElements
            Using existing system worker Init of the Elements
            Marking worker as public
            Creating new worker version
            Using new worker version {init_elements_version.id}
            ─────────────────────────────────── S3Ingest ───────────────────────────────────
            Using registry.example.com/s3-ingest:latest to provide S3Ingest
            Using existing system worker S3, Ingest of the
            Unarchiving worker
            Creating new worker version
            Using new worker version {s3_ingest_version.id}
            """
        ).strip())

    @patch("arkindex.process.management.commands.update_system_workers.parse_config", lambda: MOCK_CONFIG)
    def test_assigns_existing_compatible_versions(self):
        """
        The command should assign the feature to an existing WorkerVersion compatible with it,
        marking the version as available if necessary
        """
        worker_type = WorkerType.objects.create(slug="system", display_name="Système")
        init_elements_worker = Worker.objects.create(
            type=worker_type,
            slug="init_elements",
            name="Init of the Elements",
            public=True,
        )
        s3_ingest_worker = Worker.objects.create(
            type=worker_type,
            slug="s3_ingest",
            name="S3, Ingest of the",
            public=True,
        )

        # Those versions use the correct image, but do not have acceptable attributes
        init_elements_worker.versions.create(
            docker_image_iid="registry.example.com/init-elements:latest",
            version=1,
            configuration={"docker": {"command": "exit"}},
        )
        init_elements_worker.versions.create(
            docker_image_iid="registry.example.com/init-elements:latest",
            version=2,
            configuration={
                "user_configuration": {
                    "nobody_expects_the": {
                        "title": "required configuration",
                        "type": "int",
                        "required": True,
                    }
                },
            },
        )
        init_elements_worker.versions.create(
            docker_image_iid="registry.example.com/init-elements:latest",
            version=3,
            gpu_usage=FeatureUsage.Required,
        )
        init_elements_worker.versions.create(
            docker_image_iid="registry.example.com/init-elements:latest",
            version=4,
            model_usage=FeatureUsage.Required,
        )

        # These versions should be assigned to the features
        init_elements_version = init_elements_worker.versions.create(
            docker_image_iid="registry.example.com/init-elements:latest",
            version=5,
            state=WorkerVersionState.Available,
        )
        s3_ingest_version = s3_ingest_worker.versions.create(
            docker_image_iid="registry.example.com/s3-ingest:latest",
            version=1,
            state=WorkerVersionState.Created,
        )

        self.assertEqual(self.update_system_workers(), dedent(
            f"""
            ───────────────────────────────── InitElements ─────────────────────────────────
            Using registry.example.com/init-elements:latest to provide InitElements
            Using existing system worker Init of the Elements
            Worker is up to date
            Assigning existing worker version {init_elements_version.id} to the feature
            Using existing worker version {init_elements_version.id}
            ─────────────────────────────────── S3Ingest ───────────────────────────────────
            Using registry.example.com/s3-ingest:latest to provide S3Ingest
            Using existing system worker S3, Ingest of the
            Worker is up to date
            Assigning existing worker version {s3_ingest_version.id} to the feature
            Marking the worker version as available
            Using existing worker version {s3_ingest_version.id}
            """
        ).strip())

        self.assertEqual(WorkerType.objects.count(), 1)
        self.assertEqual(Worker.objects.count(), 2)
        self.assertEqual(WorkerVersion.objects.count(), 6)

        self.assertEqual(WorkerVersion.objects.get_by_feature(ArkindexFeature.InitElements), init_elements_version)
        self.assertEqual(WorkerVersion.objects.get_by_feature(ArkindexFeature.S3Ingest), s3_ingest_version)

        init_elements_version.refresh_from_db()
        self.check_feature_version(init_elements_version, ArkindexFeature.InitElements)
        s3_ingest_version.refresh_from_db()
        self.check_feature_version(s3_ingest_version, ArkindexFeature.S3Ingest)

    @patch("arkindex.process.management.commands.update_system_workers.parse_config", lambda: MOCK_CONFIG)
    def test_unassigns_incompatible_versions(self):
        worker_type = WorkerType.objects.create(slug="systemnt", display_name="Systemn't")
        init_elements_worker = Worker.objects.create(
            type=worker_type,
            slug="init_elements",
            name="Init of the Elements",
            public=True,
        )
        s3_ingest_worker = Worker.objects.create(
            type=worker_type,
            slug="s3_digest",
            name="S3Digest",
            public=True,
            # This isn't a system worker, so this worker being archived means the command
            # should create a new system worker instead of reusing this one
            archived=timezone.now(),
        )

        init_elements_version = init_elements_worker.versions.create(
            version=1,
            feature=ArkindexFeature.InitElements,
            docker_image_iid="registry.gitlab.teklia.com/callico/callico:latest",
            state=WorkerVersionState.Available,
            configuration={
                "docker": {
                    "command": "false",
                },
                "user_configuration": {
                    "nobody_expects_the": {
                        "title": "required configuration",
                        "type": "int",
                        "required": True,
                    },
                },
            },
        )
        # Everything is valid for this version, but it is on an archived worker
        s3_ingest_version = s3_ingest_worker.versions.create(
            feature=ArkindexFeature.S3Ingest,
            docker_image_iid="registry.example.com/s3-ingest:latest",
            version=1,
            state=WorkerVersionState.Available,
        )

        output = self.update_system_workers()

        system_worker_type = WorkerType.objects.exclude(slug=worker_type.slug).get()
        self.assertEqual(system_worker_type.slug, "system")
        self.assertEqual(system_worker_type.display_name, "System")

        self.assertEqual(Worker.objects.count(), 3)
        new_s3_ingest_worker = Worker.objects.exclude(id__in=[init_elements_worker.id, s3_ingest_worker.id]).get()
        self.assertEqual(new_s3_ingest_worker.type, system_worker_type)
        self.assertEqual(new_s3_ingest_worker.slug, "s3_ingest")
        self.assertEqual(new_s3_ingest_worker.name, "S3Ingest")

        self.assertEqual(WorkerVersion.objects.count(), 4)
        new_init_elements_version = init_elements_worker.versions.get(version=2)
        self.check_feature_version(new_init_elements_version, ArkindexFeature.InitElements)
        new_s3_ingest_version = new_s3_ingest_worker.versions.get()
        self.check_feature_version(new_s3_ingest_version, ArkindexFeature.S3Ingest)

        self.assertEqual(output, dedent(
            f"""
            ───────────────────────────────── InitElements ─────────────────────────────────
            Using registry.example.com/init-elements:latest to provide InitElements
            Current worker version: {init_elements_version.id} ({init_elements_version.docker_image_iid})
            This version has an invalid Docker image.
            This version uses a custom Docker command which could interfere with the feature.
            This version requires a custom worker configuration which could interfere with the feature.
            Unassigning feature from the current version
            Worker is up to date
            Creating new worker version
            Using new worker version {new_init_elements_version.id}
            ─────────────────────────────────── S3Ingest ───────────────────────────────────
            Using registry.example.com/s3-ingest:latest to provide S3Ingest
            Current worker version: {s3_ingest_version.id} (registry.example.com/s3-ingest:latest)
            This version is part of an archived worker.
            Unassigning feature from the current version
            Created new System worker type ({system_worker_type.id})
            Created new S3Ingest system worker
            Creating new worker version
            Using new worker version {new_s3_ingest_version.id}
            """
        ).strip())

    @patch("arkindex.process.management.commands.update_system_workers.parse_config", lambda: MOCK_CONFIG)
    def test_noop(self):
        worker_type = WorkerType.objects.create(slug="system", display_name="Système")
        worker = Worker.objects.create(type=worker_type, slug="worker", name="Worker", public=True)

        init_elements_version = worker.versions.create(
            feature=ArkindexFeature.InitElements,
            docker_image_iid="registry.example.com/init-elements:latest",
            version=1,
            state=WorkerVersionState.Available,
        )
        s3_ingest_version = worker.versions.create(
            feature=ArkindexFeature.S3Ingest,
            docker_image_iid="registry.example.com/s3-ingest:latest",
            version=2,
            state=WorkerVersionState.Available,
        )

        self.assertEqual(self.update_system_workers(), dedent(
            f"""
            ───────────────────────────────── InitElements ─────────────────────────────────
            Using registry.example.com/init-elements:latest to provide InitElements
            Current worker version: {init_elements_version.id} (registry.example.com/init-elements:latest)
            Worker is up to date
            Worker version for InitElements is up to date
            ─────────────────────────────────── S3Ingest ───────────────────────────────────
            Using registry.example.com/s3-ingest:latest to provide S3Ingest
            Current worker version: {s3_ingest_version.id} (registry.example.com/s3-ingest:latest)
            Worker is up to date
            Worker version for S3Ingest is up to date
            """
        ).strip())

        self.assertEqual(WorkerType.objects.count(), 1)
        self.assertEqual(Worker.objects.count(), 1)
        self.assertEqual(WorkerVersion.objects.count(), 2)

        self.assertEqual(WorkerVersion.objects.get_by_feature(ArkindexFeature.InitElements), init_elements_version)
        self.assertEqual(WorkerVersion.objects.get_by_feature(ArkindexFeature.S3Ingest), s3_ingest_version)

        init_elements_version.refresh_from_db()
        self.check_feature_version(init_elements_version, ArkindexFeature.InitElements)
        s3_ingest_version.refresh_from_db()
        self.check_feature_version(s3_ingest_version, ArkindexFeature.S3Ingest)
