from unittest.mock import call, patch

from django.urls import reverse
from rest_framework import status

from arkindex.ponos.models import Farm
from arkindex.process.models import FeatureUsage, ProcessMode, Worker, WorkerVersion, WorkerVersionState
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role, User


class TestWorkerVersions(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.worker_reco = Worker.objects.get(slug="reco")
        cls.worker_dla = Worker.objects.get(slug="dla")

        cls.version_1 = cls.worker_reco.versions.get()
        cls.version_2 = cls.worker_dla.versions.get()

        farm = Farm.objects.first()
        process = cls.corpus.processes.create(
            mode=ProcessMode.Workers,
            creator=cls.user,
            farm=farm,
            name="test_process"
        )
        process.run()
        cls.task = process.tasks.get()

    # Tests on get_query_set for WorkerVersionList
    def test_list_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list(self):
        self.client.force_login(self.user)
        last_version = self.worker_reco.versions.create(
            configuration={"a": "bc"},
            state=WorkerVersionState.Error,
            revision_url="https://gitlab.com/NERV/eva/commit/63e377e7f88c743d8428fc4e4eaedfc1c9356754"
        )
        last_version.created = "2050-09-09T09:09:09.090909Z"
        last_version.save()

        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 2,
            "next": None,
            "number": 1,
            "previous": None,
            "results": [
                {
                    "id": str(last_version.id),
                    "configuration": {"a": "bc"},
                    "docker_image_iid": last_version.docker_image_iid,
                    "state": "error",
                    "gpu_usage": FeatureUsage.Disabled.value,
                    "model_usage": FeatureUsage.Disabled.value,
                    "worker": {
                        "id": str(self.worker_reco.id),
                        "name": self.worker_reco.name,
                        "type": self.worker_reco.type.slug,
                        "slug": self.worker_reco.slug,
                        "description": self.worker_reco.description,
                        "archived": bool(self.worker_reco.archived),
                        "repository_url": self.worker_reco.repository_url,
                    },
                    "version": None,
                    "tag": None,
                    "branch": None,
                    "revision_url": "https://gitlab.com/NERV/eva/commit/63e377e7f88c743d8428fc4e4eaedfc1c9356754",
                    "created": "2050-09-09T09:09:09.090909Z",
                },
                {
                    "id": str(self.version_1.id),
                    "configuration": {"test": 42},
                    "docker_image_iid": self.version_1.docker_image_iid,
                    "state": "available",
                    "gpu_usage": FeatureUsage.Disabled.value,
                    "model_usage": FeatureUsage.Disabled.value,
                    "worker": {
                        "id": str(self.worker_reco.id),
                        "name": self.worker_reco.name,
                        "type": self.worker_reco.type.slug,
                        "slug": self.worker_reco.slug,
                        "description": self.worker_reco.description,
                        "archived": bool(self.worker_reco.archived),
                        "repository_url": self.worker_reco.repository_url,
                    },
                    "version": 1,
                    "tag": None,
                    "branch": None,
                    "revision_url": None,
                    "created": self.version_1.created.isoformat().replace("+00:00", "Z"),
                }
            ]
        })

    @patch("arkindex.users.utils.get_max_level", return_value=Role.Guest.value)
    def test_list_requires_contributor(self, max_level_mock):
        """
        User is not able to list worker versions with a guest access
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {
            "detail": "You do not have an execution access to this worker.",
        })
        self.assertEqual(max_level_mock.call_count, 1)
        self.assertEqual(max_level_mock.call_args_list, [
            call(self.user, self.worker_reco)
        ])

    def test_list_public_worker(self):
        """
        Any authenticated user is able to list versions for public workers
        """
        user = User.objects.create_user("user42@test.fr", "abcd")
        user.verified_email = True
        user.save()
        self.worker_reco.public = True
        self.worker_reco.save()
        self.client.force_login(user)

        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertEqual(data["count"], 1)
        self.assertEqual(data["results"][0]["id"], str(self.worker_reco.versions.first().id))

    def test_list_filter_worker(self):
        self.client.force_login(self.user)
        worker_2 = Worker.objects.create(
            name="Worker 2",
            slug="worker_2",
            type=self.worker_dla.type,
        )
        version_2 = WorkerVersion.objects.create(
            worker=worker_2,
            version=1,
            configuration={"test": "test2"},
            gpu_usage=FeatureUsage.Disabled
        )

        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:worker-versions", kwargs={"pk": str(worker_2.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertEqual(len(data["results"]), 1)
        version = data["results"][0]
        self.assertEqual(version["id"], str(version_2.id))

    def test_list_simple_mode(self):
        """
        With the mode attribute set to simple, worker versions with no tag are excluded
        """
        WorkerVersion.objects.bulk_create([
            WorkerVersion(
                worker=self.worker_reco,
                configuration={"test": "42"},
                tag=tag,
                version=i,
            ) for i, tag in enumerate((None, "v0.1", "test"), start=2)
        ])

        with self.subTest(mode="complete"), self.assertNumQueries(4):
            response = self.client.get(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
                {"mode": "complete"},
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
            data = response.json()
            self.assertEqual(data["count"], 4)
            self.assertCountEqual(
                [version["tag"] for version in data["results"]],
                ["v0.1", "test", None, None],
            )

        # Simple mode filters out versions without a tag
        with self.subTest(mode="simple"), self.assertNumQueries(4):
            response = self.client.get(
                reverse("api:worker-versions", kwargs={"pk": str(self.worker_reco.id)}),
                {"mode": "simple"},
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
            data = response.json()
            self.assertEqual(data["count"], 2)
            self.assertCountEqual(
                [version["tag"] for version in data["results"]],
                ["v0.1", "test"],
            )
